// Neon++ Library, (C) 2006 M. Derezynski
// A C++ wrapper library for neon <http://www.webdav.org/neon>, (C) Joe Orton

#include <iostream>
#include <istream>
#include <ostream>
#include <string>

#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include "session.hh" 
#include "request.hh" 

namespace Neon
{
  Request::~Request ()
  {
    if (m_request) ne_request_destroy (m_request);
    if (m_session) delete m_session;
  }

  Request::Request (std::string const& hostname,
                    std::string const& path,
                    Neon::HttpPort port,
                    ResponseRead read,
                    Method method,
                    std::string const& scheme)

  throw (Neon::Session::SessionCreateError, InvalidRequestError) 

            : m_request (0),
              m_session (0),
              m_dispatched (false),
              m_read (read),
              m_method (method),
              m_handle_redirect (false)

  {
    m_session = new Neon::Session (hostname, scheme, port);
    m_request = ne_request_create (m_session->m_session, (m_method == Neon::Request::METHOD_GET) ? "GET" : "POST", path.c_str());

    if (!m_request)
      throw InvalidRequestError (ne_get_error (m_session->m_session));

    if (m_read == Neon::Request::RESPONSE_READ)
    {
      ne_add_response_body_reader (m_request, ne_accept_2xx, read_block, &m_vec);
    }
  }

  Neon::Session&
  Request::get_session ()
  {
    return *m_session;
  }

  void
  Request::handle_redirect (bool handle)
  {
    m_handle_redirect = handle;
  }

  void 
  Request::set_request_data (const char * data, size_t data_size)
  {
    ne_set_request_body_buffer (m_request, data, data_size); 
  }

  const ne_status*
  Request::get_status () const
  {
    return ne_get_status (m_request);
  }

  int
  Request::read_block (void * udata, const char * data, size_t len)
  {
    std::vector<unsigned char> *vec = (std::vector<unsigned char>*)(udata);
    for (unsigned int n = 0; n < len; ++n)
      vec->push_back ((unsigned char)data[n]);
    return 0;
  }

  void
  Request::clear ()
  {
    m_dispatched = false;
    m_vec.clear(); 
  }

#ifdef NEONPP_EXCEPTIONS_ENABLED
  void
#else
  int
#endif
  Request::dispatch ()
  {
    if (m_dispatched) return NE_OK;

    int result = ne_request_dispatch (m_request);
    int status = ne_get_status (m_request)->code;

    if ((status == 302) && m_handle_redirect)
      {
        const char *redir = ne_get_response_header (m_request, "Location");

        if (redir != NULL)
          {
            ne_request_destroy (m_request);
            m_request = 0;
            delete m_session;
            m_session = 0;

            ne_uri uri;

            if (ne_uri_parse (redir, &uri))
              return NE_ERROR; 

            std::string path;
            path = uri.path;

#ifdef HAVE_NEON_026
            if (uri.query)
              {
                path.append ("?");
                path.append (uri.query);
              }
            if (uri.fragment)
              {
                path.append ("#");
                path.append (uri.fragment);
              }
#endif
            m_session = new Neon::Session (uri.host, uri.scheme, uri.port);
            m_request = ne_request_create (m_session->m_session, (m_method == Neon::Request::METHOD_GET) ? "GET" : "POST", path.c_str());

            if (!m_request)
              throw InvalidRequestError (ne_get_error (m_session->m_session));

            if (m_read == Neon::Request::RESPONSE_READ)
              {
                ne_add_response_body_reader (m_request, ne_accept_2xx, read_block, &m_vec);
              }

            dispatch ();
          }
      }

    m_dispatched = true;

#ifdef NEONPP_EXCEPTIONS_ENABLED
    switch (result)
    {
      case NE_OK:
        return;

      case NE_AUTH:
        throw NeAuthError (ne_get_error (ne_get_session (m_request)));

      case NE_PROXYAUTH:
        throw NeProxyAuthError (ne_get_error (ne_get_session (m_request)));

      case NE_CONNECT:  
        throw NeConnectError (ne_get_error (ne_get_session (m_request)));

      case NE_ERROR:
        throw NeGenericError (ne_get_error (ne_get_session (m_request)));
    }
#else
    return result;
#endif

  }

  void
  Request::add_request_header (std::string const& name, std::string const& value)
  {
    ne_add_request_header (m_request, name.c_str(), value.c_str());
  }

  void
  Request::add_request_header (Neon::Request::HeaderType type)
  {
    switch (type)
    {
      case HEADER_WWWFORM:
        ne_add_request_header (m_request, "Content-Type", "application/x-www-form-urlencoded");
        break;

      case HEADER_TEXTXML:
        ne_add_request_header (m_request, "Content-Type", "text/xml");
        break;
    }
  }

  std::ostream&
  operator<<(std::ostream& os, Request & r)
  {
    r.dispatch ();
    os.write (reinterpret_cast<const char*>(&(r.m_vec[0])), r.m_vec.size());
    return os;
  }

  std::ostream&
  Request::operator<<(std::ostream& os)
  {
    dispatch ();
    os.write (reinterpret_cast<const char*>(&(m_vec[0])), m_vec.size());
    return os;
  }

  void
  Request::operator>> (std::string& out)
  {
    dispatch ();
    if (m_vec.size())
      {
        out.append (reinterpret_cast<const char*>(&m_vec[0])); 
        out.append ('\0');
      }
  }

  const unsigned char *
  Request::get_data ()
  {
    return &m_vec[0];
  }

  size_t
  Request::get_data_size ()
  {
    return m_vec.size();
  }

  std::string
  Request::get_error () const
  {
    return std::string (ne_get_error (ne_get_session (m_request)));
  }

  void
  Request::get_response_headers (Neon::Request::ResponseHeaders & headers)
  {
#ifdef NEONPP_EXCEPTIONS_ENABLED
    if (!m_request) 
        throw (NoCurrentRequestError);
#else
    if (!m_request)
        return;
#endif

    const char *name = 0, *value = 0;
    void* cursor = NULL;
    ne_response_header_iterate (m_request, cursor, &name, &value);
    while (cursor = ne_response_header_iterate (m_request, cursor, &name, &value))     
    {
      if (name && value) headers.insert (std::make_pair (std::string (name), std::string (value)));
    }
  }

}
